#include "tcpip.h"
#include <errno.h>

struct realm *icmp_realm;

struct icmp_head *icmprules=NULL;
unsigned int num_icmp_head=0;
unsigned int num_icmp_rules=0;

void icmp_match(struct packet *p, unsigned int l)
{
	struct icmp_head *h;
	struct pkt_iphdr *iph;
	struct pkt_icmphdr *icmph;

	iph=p->layer[l].h.ip;
	icmph=p->layer[l+1].h.icmp;

	for(h=icmprules; h; h=h->next) {
		if ( 
			(((h->saddr&h->smask) == (iph->saddr&h->smask))
			^ ((h->flags&NEGATE_SADDR)!=0) ) &&
			
			(((h->daddr&h->dmask) == (iph->daddr&h->dmask))
			^ ((h->flags&NEGATE_DADDR)!=0) ) ) {

			if ( h->flags&CHECK_ITYPE &&
				!((h->itype==icmph->type)^
				((h->flags&NEGATE_ITYPE)!=0)) ) continue;

			if ( h->flags&CHECK_ICODE &&
				!((h->icode==icmph->code)^
				((h->flags&NEGATE_ICODE)!=0)) ) continue;

			tcpip_traverse(p, l, h->rules);
		}
	}
}

struct icmp_head *icmp_rlink(struct icmp_head *n)
{
	struct icmp_head *tmp;

	for(tmp=icmprules; tmp; tmp=tmp->next) {
		if ( n->saddr==tmp->saddr && n->smask==tmp->smask &&
		  	n->daddr==tmp->daddr && n->dmask==tmp->dmask &&
		  	n->itype==tmp->itype && n->icode==tmp->icode &&
		  	n->flags==tmp->flags ) {
			free(n);
			return tmp;
		}
	}

	/* Link this node in with the list */
	num_icmp_head++;
	n->next=icmprules;
	icmprules=n;
	return n;
}

int icmp_commit(struct rule *r)
{
	struct icmp_head *hd;
	struct ip_rule *rule;
	struct criteria *c;
	int err=0;

	if ( !(hd=calloc(1, sizeof(*hd))) ) {
		return 0;
	}

	if ( !(rule=calloc(1, sizeof(*rule))) ) {
		free(hd);
		return 0;
	}

	for(c=r->criteria; c; c=c->next) {
		if ( !strcmp("src", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_SADDR;
			if ( !parse_ipaddr(c->args, &hd->saddr, &hd->smask) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("dst", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_DADDR;
			if ( !parse_ipaddr(c->args, &hd->daddr, &hd->dmask) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("itype", c->crit) ) {
			unsigned long int num;
			if ( c->negate ) hd->flags|=NEGATE_ITYPE;
			hd->flags|=CHECK_ITYPE;

			if ( (num=strtoul(c->args, NULL, 10))==ULONG_MAX &&
				errno==ERANGE ) {
				err=1;
				break;
			}

			if ( num&(~0xff) ) {
				err=1;
				break;
			}

			hd->itype=(u_int8_t)num&0xff;
			continue;
		}

		if ( !strcmp("icode", c->crit) ) {
			unsigned long int num;

			if ( c->negate ) hd->flags|=NEGATE_ICODE;
			hd->flags|=CHECK_ICODE;

			if ( (num=strtoul(c->args, NULL, 10))==ULONG_MAX &&
				errno==ERANGE ) {
				err=1;
				break;
			}

			if ( num&(~0xff) ) {
				err=1;
				break;
			}

			hd->icode=(u_int8_t)num&0xff;
			continue;
		}

		if ( !iprule_addcrit(c, rule, r->modifier) ) {
			err=1;
			break;
		}
	}

	if ( err ) {
		iprule_cleanup(rule);
		free(hd);
		return 0;
	}

	/* Can't fail */
	hd=icmp_rlink(hd);
	
	memcpy(&rule->a, &r->alert, sizeof(rule->a));

	/* Link rule on to rule header */
	rule->next=hd->rules;
	hd->rules=rule;
	num_icmp_rules++;
	
	return 1;
}

struct proto icmp_p=init_proto_ex("icmp", icmp_decode, icmp_dprint);
struct proto_req icmp_r[]={
	proto_request("ipv4", 1),
	null_request()
};

int icmp_dprint(struct layer *l, char *buf, int buflen)
{
	return snprintf(buf, buflen,
		"TYPE=%u CODE=%u",
		l->h.icmp->type,
		l->h.icmp->code);
}

int icmp_decode(struct packet *p)
{
	struct proto_child *pc;
	struct layer *l=&p->layer[p->llen-1];
	unsigned long icmpid;
	
	if ( p->llen >= PKT_LAYERS ) return 0;
	
	if ( (p->layer[p->llen].h.raw=
		l->h.raw+sizeof(pkt_icmphdr_t)) 
		>= p->end ) goto finish;

	/* Might look odd but we just combine the type
	 * and code in order to make an ID for a child
	 * protocol */
	icmpid=(l->h.icmp->type<<8 | l->h.icmp->code);
	
	for(pc=l->proto->children; pc; pc=pc->next)
	{
		if ( icmpid == pc->id ) {
			p->layer[p->llen++].proto=pc->proto;
			return pc->proto->decode(p);
		}
	}

finish:
	p->layer[p->llen++].proto=NULL;
	return 0;
}
